/* ifile - intelligent mail filter for EXMH/MH
   Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>
   
   ifile Copyright (C) 1997 by Jason Rennie <jr6b+@andrew.cmu.edu>
   ifilter.c code Copyright (C) 1997 by Jason Rennie <jr6b+@andrew.cmu.edu>
                                 and Diego Zamboni <zamboni@cs.purdue.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/*
  ifilter.c
  
  C implementation of ifilter.perl, written by Jason Rennie.
  
  C code primarily written by Diego Zamboni <zamboni@cs.purdue.edu>, Mar 7, 1997
     additions and changes by Jason Rennie <jr6b@cs.cmu.edu>

  No attempt at modification of the algorithm of program was
  made. This reimplementation was made solely with the attempt of
  making the program run faster. I receive a lot of mail, and
  filtering it just takes too much time.
  
  Some optimizations were made, particularily in the analysis of the
  message. Some of the most important ones are:
  
  - The message is "pre-partitioned" in words and the words are stored
    in an array which is then used repeatedly for computing the target
    value for each folder. This is _much_ faster than repartitioning
    the entire message for each category.

  I am still thinking about some optimizations that could be made to
  the database-reading part. But for now, the filtering time is down
  to ~3 sec, from ~17 sec previously. Also, the memory footprint of
  this program is _much_ smaller than the perl one, which makes its
  execution faser and less noticeable (my machine slowed down every
  time exmh did an inc with the perl version).

  Cheers to Jason for writing such a useful program. I hope this can
  be useful too.

  */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>     // wait()
#include <string.h>
#include <math.h>
#include <unistd.h>       // pipe() fork() close() dup2() execlp()
#include <ctype.h>        // tolower()
#include "extendable_array.h"
#include "assoc_array_int.h"

/* rcvstore path --configure to fit your system. */
/* this is configured in the install.perl script.  Do not change manually! */
#define RCVSTORE "/usr/lib/mh/rcvstore"

#ifndef CLOCKS_PER_SECOND
  #ifdef CLOCKS_PER_SEC
  #define CLOCKS_PER_SECOND CLOCKS_PER_SEC
  #else
    #ifdef CLK_TCK
    #define CLOCKS_PER_SECOND CLK_TCK
    #else
    #define CLOCKS_PER_SECOND 100
    #endif
  #endif
#endif

FILE *TMP;

/* Read a line, stripping the feedline '\n' */
/* void readline(char *s, int n, FILE *stream) { */
#define readline(s,n,stream) {\
  fgets(s, n, stream);\
    if(s[strlen(s)-1]=='\n') {\
    s[strlen(s)-1]='\0';\
  }\
}

/* Complain and die */
void ouch(char *msg)
{
  fprintf(stderr, "%s\n",msg);
  fprintf(TMP, "%s\n",msg);
  exit(1);
}

/* Pass message to rcvstore */
/* Returns 0 upon safe execution, -1 or rcvstore exit status otherwise */
int rcvstore(extendable_array lines, char *mbox)
{
  int fifo[2], status;  /* Pipe file descriptors */
  char *mboxplus;
  int i;
  pid_t child_pid;
  char *buf;
  int buf_siz;
  
  mboxplus=(char*)malloc(strlen(mbox)+2);
  strcpy(mboxplus,"+");
  strcat(mboxplus, mbox);
  
  fprintf(TMP, "Rcvstoring in %s\n", mbox);
  fprintf(stderr, "Rcvstoring in %s\n", mbox);
  
  if(pipe(fifo)==-1)
    {
      ouch("Could not open pipe to rcvstore.");
    }
  
  switch(child_pid = fork())
    {
    case -1:   /* Error */
      ouch("Fork to rcvstore failed.\n");
      exit(2);
    case 0:    /* Child */
      if(close(fifo[1])==-1)   /* We don't need the writing end */
	ouch("Close of writing end in child failed.");
      dup2(fifo[0], 0);  /* Make stdin refer to the inherited reading
			    end of the pipe. */
      execlp(RCVSTORE, RCVSTORE, mboxplus, NULL);
      ouch("execlp to rcvstore failed.\n");
    default:   /* Parent */
      if(close(fifo[0])==-1)  /* We don't need the reading end */
	ouch("Close of reading end in parent failed.");
      /* pipe message to rcvstore child */
      for(i=0; i<lines.elems; i++)
	{
	  /* send line plus '\n', malloc extra byte for \0 */
	  buf_siz = strlen(EXT_ARRAY_GET(lines, char *, i)) + 1;
	  buf = malloc(buf_siz+1);
	  strcpy(buf, EXT_ARRAY_GET(lines,char*,i));
	  strcat(buf, "\n");
	  write(fifo[1], buf, buf_siz);
	  free(buf);
	}
      if (close(fifo[1])==-1)
	ouch("Close of writing end in parent failed.");
      /* allows exmh to recognize new message, highlight proper folder */
      fprintf(TMP, "Waiting for rcvstore fork to return\n");
      if (waitpid(child_pid, &status, 0) == child_pid && WIFEXITED(status))
	return WEXITSTATUS(status);
      return -1;
    }
}

/* Main program */
int main(int argc, char **argv) {
  /* To the extent that it was possible, the original variable names
     were kept to avoid confusion. */
  clock_t DMZ_start, DMZ_end;
  int i,j;
  char *home_dir = NULL;
  char *mh_profile = NULL;
  FILE *MH = NULL;
  char *mail_path = NULL;
  char tmpstr[BUFSIZ*8], *tmpstrptr = NULL;
  char *draft_folder = NULL;
  char *acc_file = NULL;
  FILE *ACC = NULL;
  int filters, refiles;
  float accuracy = 0.0;
  int tot_words;
  extendable_array tv;
  char *data_file = NULL;
  FILE *DATA = NULL;
  int tvsize;
  assoc_array_int n, docs, index, age, nk;
  int tot_docs;
  int tmpint;
  extendable_array word;
  char *curword = NULL;
  int cat_num, freq;
  char tmpstr2[BUFSIZ];
  float best_rating;
  int best_target;
  extendable_array lines;
  float target_rating;
  char nonalpha[256];
  char *wrd = NULL;
  extendable_array words;
  int word_count;
  float nval, docval;
  
  /* Get initial time */
  DMZ_start=clock();
  
  /* Open log file and write initial info */
  TMP=fopen("/tmp/ifilter.info", "a");
  if(TMP==NULL)
    {
      fprintf(stderr, "Can't open /tmp/ifilter.info\n");
    }
  
  fprintf(TMP, "ifilter (C version) called (");
  for(i=0; i<argc; i++)
    {
      fprintf(TMP, "%s", argv[i]);
      if (i<argc-1) fprintf(TMP, " ");
    }
  fprintf(TMP, ")\n");
  
  /* Get home directory */
  home_dir=getenv("HOME");
  if(home_dir==NULL)
    {
      fprintf(stderr, "Fatal: HOME environment variable not defined!\n");
      fclose(TMP);
      exit(1);
    }
  
  /* Find mail dir in .mh_profile */
  mh_profile=(char*)malloc(strlen(home_dir)+strlen("/.mh_profile")+1);
  strcpy(mh_profile, home_dir);
  strcat(mh_profile, "/.mh_profile");
  MH=fopen(mh_profile, "r");
  /* If .mh_profile is not found, set default Mail directory */
  if(MH==NULL)
    {
      mail_path="Mail/";
    } else {
      while(!feof(MH)) {
	readline(tmpstr, sizeof(tmpstr), MH);
	tmpstrptr=strtok(tmpstr, " \t");
	if(strcmp(tmpstrptr, "Path:")==0)
	  {
	    tmpstrptr=strtok(NULL, " \t");
	    mail_path=(char*)malloc(strlen(tmpstrptr)+2);
	    strcpy(mail_path, tmpstrptr);
	    strcat(mail_path, "/");
	  } 
	else if(strcmp(tmpstrptr, "draft-folder:")==0)
	  {
	    tmpstrptr=strtok(NULL, " \t");
	    draft_folder=(char*)malloc(strlen(tmpstrptr)+1);
	    strcpy(draft_folder, tmpstrptr);
	  }
      }
      fclose(MH);
    }
  
  /* Update accuracies file */
  acc_file=(char*)malloc(strlen(home_dir)+strlen("/.idata_accuracy")+1);
  strcpy(acc_file, home_dir);
  strcat(acc_file, "/.idata_accuracy");
  filters=1;
  refiles=0;
  
  ACC=fopen(acc_file, "r");
  if(ACC!=NULL)
    {
      readline(tmpstr, sizeof(tmpstr), ACC);
      tmpstrptr=strtok(tmpstr, " \t=");
      if(strcmp(tmpstrptr, "filters")==0)
	{
	  filters+=atoi(strtok(NULL, " \t="));
	  tmpstrptr=strtok(NULL, " \t=");
	  if(strcmp(tmpstrptr, "refiles")==0)
	    {
	      tmpstrptr=strtok(NULL, " \t=");
	      refiles+=atoi(tmpstrptr);
	    }
	}
      accuracy=(((float)(filters-refiles)/(float)filters)*10000.0)/100.0;
      fprintf(TMP, "filters = %d  refiles = %d\n", filters, refiles);
      fprintf(TMP, "Accuracy = %.2f %%\n", accuracy);
      
      fclose(ACC);
    }
  
  ACC=fopen(acc_file, "w");
  if(ACC!=NULL)
    {
      fprintf(ACC, "filters = %d  refiles = %d\n", filters,
	      refiles);
      fprintf(ACC, "Accuracy = %.2f %%\n", accuracy);
      fclose(ACC);
    } else {
      fprintf(TMP, "Was not able to write accuracies file\n");
    }
  
  /* Now read the database */
  tot_words=0;
  
  EXT_ARRAY_INIT(tv,char*,EXT_ARRAY_CHUNK_SIZE);
  EXT_ARRAY_SET(tv,char*,0,NULL);
  tvsize=0;
  
  data_file=(char*)malloc(strlen(home_dir)+strlen("/.idata")+1);
  strcpy(data_file, home_dir);
  strcat(data_file, "/.idata");
  
  DATA=fopen(data_file, "r");
  if(DATA!=NULL)
    {
      /* Read the category names */
      readline(tmpstr, sizeof(tmpstr), DATA);
      tmpstrptr=strtok(tmpstr, " \t");
      while(tmpstrptr!=NULL)
	{
	  tvsize++;
	  EXT_ARRAY_SET(tv,char*,tvsize,strdup(tmpstrptr));
	  tmpstrptr=strtok(NULL, " \t");
	}

      /* Read total number of words for each category */
      readline(tmpstr, sizeof(tmpstr), DATA);
      assoc_array_int_init(n);
      i=1;
      tmpstrptr=strtok(tmpstr, " \t");
      while(tmpstrptr!=NULL)
	{
	  assoc_array_int_set(n, EXT_ARRAY_GET(tv,char*,i),
			      atoi(tmpstrptr));
	  i++;
	  tmpstrptr=strtok(NULL, " \t");
	}
      if(i<=tvsize)
	{
	  ouch("Bad data file format - line #2");
	}
      
      /* Read number of documents for each category */
      readline(tmpstr, sizeof(tmpstr), DATA);
      assoc_array_int_init(docs);
      tot_docs=0;
      i=1;
      tmpstrptr=strtok(tmpstr, " \t");
      while(tmpstrptr!=NULL)
	{
	  tmpint=atoi(tmpstrptr);
	  assoc_array_int_set(docs, EXT_ARRAY_GET(tv,char*,i),tmpint);
	  i++;
	  tot_docs+=tmpint;
	  tmpstrptr=strtok(NULL, " \t");
	}
      if(i<=tvsize)
	{
	  ouch("Bad data file format - line #3");
	}
      
      /* Read word frequencies */
      EXT_ARRAY_INIT(word,char*,8000);
      EXT_ARRAY_SET(word,char*,0,NULL);
      assoc_array_int_init(index);
      assoc_array_int_init(age);
      assoc_array_int_init(nk);
      
      readline(tmpstr, sizeof(tmpstr), DATA);
      while(!feof(DATA))
	{
	  tmpstrptr=strtok(tmpstr, " \t");
	  if(tmpstrptr==NULL)
	    break;
	  tot_words++;
	  EXT_ARRAY_SET(word,char*,tot_words,strdup(tmpstrptr));
	  assoc_array_int_set(index, tmpstrptr, tot_words);
	  curword=tmpstrptr;
	  
	  /* Read age of word */
	  tmpstrptr=strtok(NULL, " \t");
	  assoc_array_int_set(age, curword, atoi(tmpstrptr));
	  
	  /* Read frequency per category cat_num:freq */
	  tmpstrptr=strtok(NULL, ":");
	  while(tmpstrptr!=NULL)
	    {
	      cat_num=atoi(tmpstrptr);
	      tmpstrptr=strtok(NULL, " \t");
	      freq=atoi(tmpstrptr);
	      sprintf(tmpstr2, "%s,%s",
		      EXT_ARRAY_GET(word,char*,tot_words),
		      EXT_ARRAY_GET(tv,char*,cat_num));
	      assoc_array_int_set(nk, tmpstr2, freq);
	      
	      tmpstrptr=strtok(NULL, ":");
	    }
	  readline(tmpstr, sizeof(tmpstr), DATA);
	}
    } else {
      /* If no .idata is found, just pass to rcvstore */
      EXT_ARRAY_INIT(lines,char*,EXT_ARRAY_CHUNK_SIZE);
      i=0;
      
      /* Read the message in an array for processing */
      readline(tmpstr, sizeof(tmpstr), stdin);
      while(!feof(stdin))
	{
	  EXT_ARRAY_SET(lines,char*,i,tmpstr);
	  i++;
	  readline(tmpstr, sizeof(tmpstr), stdin);
	}
      
      rcvstore(lines, "inbox");
      exit(0);
    }
  
#ifdef DEBUG
  fprintf(stderr,"Finishing reading database.\n");
  DMZ_end=clock();
  fprintf(stderr,"Time elapsed so far: %.3f sec\n",
	  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
#endif

  fclose(DATA);
  
  best_rating=0.0;
  best_target=0;
  
  /* Initialize the non-alphabetic-character token separator */
  j=0;
  for(i=1; i<=255; i++)
    {
      if(!((i>='a' && i<='z') || (i>='A' && i<='Z')))
	nonalpha[j++]=i;
    }
  nonalpha[j]='\0';
  
  EXT_ARRAY_INIT(lines,char*,EXT_ARRAY_CHUNK_SIZE);
  EXT_ARRAY_INIT(words,char*,500);
  i=0;
  word_count=0;
  
  /* Read the message in an array for processing */
  readline(tmpstr, sizeof(tmpstr), stdin);
  while(!feof(stdin))
    {
      EXT_ARRAY_SET(lines,char*,i,strdup(tmpstr));
      i++;
      /* Split line in words and store them in another array */
      tmpstrptr=strtok(tmpstr,nonalpha);
      while(tmpstrptr!=NULL)
	{
	  wrd=tmpstrptr;
	  /* Skip tiny words */
	  if(strlen(wrd)>2)
	    {
	      /* Convert to lower case */
	      while((*wrd)!='\0')
		{
		  (*wrd)=tolower(*wrd);
		  wrd++;
		}
	      EXT_ARRAY_SET(words,char*,word_count,strdup(tmpstrptr));
	      word_count++;
	    }
	  tmpstrptr=strtok(NULL,nonalpha);
	}
      readline(tmpstr, sizeof(tmpstr), stdin);
    }

#ifdef DEBUG
  fprintf(stderr,"Finishing reading message.\n");
  DMZ_end=clock();
  fprintf(stderr,"Time elapsed so far: %.3f sec\n",
	  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
#endif
  
  docval=(float)(tot_docs+tvsize);
  
  /* Now do the analysis */
  for(i=1; i<=tvsize; i++)
    {
      target_rating=0.0;
      tmpstrptr=EXT_ARRAY_GET(tv,char*,i);
      nval=(float)(assoc_array_int_get(n,tmpstrptr)+tot_words);
      for(j=0; j<word_count; j++)
	{
	  wrd=EXT_ARRAY_GET(words,char*,j);
	  sprintf(tmpstr2, "%s,%s", wrd, tmpstrptr);
	  target_rating+=log(((float)(assoc_array_int_get(nk,tmpstr2)+1))/
			     nval);
	}
      
#ifdef VERBOSE
      fprintf(TMP,"%s given rating of %f\n", tmpstrptr,
	      target_rating);
#endif
      target_rating+=log((float)(assoc_array_int_get(docs,tmpstrptr)+1)/
			 docval);
      if((target_rating>best_rating) || (best_target==0))
	{
	  best_rating=target_rating;
	  best_target=i;
	}
    }
  
  fprintf(TMP, "Filed message in %s (%d)\n",
	  EXT_ARRAY_GET(tv,char*,best_target), best_target);
  
  /* Compute elapse time */
  DMZ_end=clock();
  fprintf(TMP, "That took %.3f seconds.\n",
	  ((double)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
#ifdef DEBUG
  fprintf(stderr, "That took %.3f seconds.\n",
	  ((double)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
#endif
  
  fprintf(TMP, "\n");
  fclose(TMP);
  
  rcvstore(lines, EXT_ARRAY_GET(tv,char*,best_target));
  
  exit(0);
}
